import re
from collections import OrderedDict
from dektools.dict import assign


class ImportsException(Exception):
    pass


class Imports:
    def __init__(self, parents=None, data=None):
        self.parents = parents or []
        self.data = data or {}
        self.data_formatted = {}
        self.data_primary = OrderedDict()
        self.data_shortcut = {}
        self.update(self.data)

    def derive(self):
        return self.__class__([self])

    def dependency(self):
        self.parents = []

    def add_shortcut(self, typed, key, value):
        self.data_shortcut.setdefault(typed, {})[key] = value

    def add_primary(self, key, value):
        self.data_primary[key] = value

    def add_item(self, name, data):
        self.data[name] = assign(self.data.get(name) or {}, data)
        last_name = None
        for x in reversed(name.split('.')):
            if last_name is None:
                label = x
            else:
                label = f'{x}.{last_name}'
            last_name = label
            self.data_formatted.setdefault(label, set()).add(name)

    def update(self, data):
        for k, v in data.items():
            self.add_item(k, v)

    def get_item(self, key, bases=None):
        bases = bases or ()
        if not isinstance(bases, (list, tuple)):
            bases = [bases]
        label_set = self.get_label_formatted_set(key)
        self.validate(label_set, key, True)
        label_primary = self.get_label_primary(key)
        if label_primary and label_primary in label_set:
            return label_primary
        if len(label_set) > 1:
            label_set = self.checkout_hits(label_set, bases)
            self.validate(label_set, key, False)
        if key in label_set:
            return key
        return next(iter(label_set))

    @staticmethod
    def validate(label_set, key, allow_multiple):
        if not label_set:
            raise ImportsException(f'Can not find module by [ {key} ]')
        elif len(label_set) > 1:
            if key not in label_set:
                if not allow_multiple:
                    raise ImportsException(f'Check out multiple module [{label_set}] by [ {key} ]')

    def get_label_formatted_set(self, key):
        if key in self.data_formatted:
            return self.data_formatted[key]
        for parent in self.parents:
            value = parent.get_label_formatted_set(key)
            if value is not None:
                return value

    def get_key_by_shortcut(self, key, typed):
        mt = self.data_shortcut.get(typed)
        if mt:
            return mt.get(key)
        for parent in self.parents:
            value = parent.get_key_by_shortcut(key, typed)
            if value is not None:
                return value

    def get_label_primary(self, key):
        for query, label in self.data_primary.items():
            m = re.search(query, key)
            if m:
                label_primary = label.format(*m.groups())
                if self.get_label_data(label_primary):
                    return label_primary
        for parent in self.parents:
            value = parent.get_label_primary(key)
            if value is not None:
                return value

    def get_label_data(self, label):
        if label in self.data:
            return self.data[label]
        for parent in self.parents:
            value = parent.get_label_data(label)
            if value is not None:
                return value

    def checkout_hits(self, label_set, bases):
        result = set()
        for label in label_set:
            for base in bases:
                if not self.is_hit(label, base):
                    break
            else:
                result.add(label)
        return result

    def is_hit(self, label, key):
        data = self.get_label_data(label)
        if data:
            value = data.get('value')
            bases = data.get('bases') or []
            lists = ([value] if value else []) + bases
            for item in lists:
                if key == item or item.endswith(f'.{key}'):
                    return True
            for item in lists:
                if self.is_hit(item, key):
                    return True
        return False


class ImportsStatement:
    def __init__(self, imports: Imports):
        self.imports = imports
        self.statements = {}

    def as_statements(self):
        return '\n'.join(sorted(self.statements.values()))

    def var(self, module_guss, bases=None, alias=None):
        module = self.imports.get_item(module_guss, bases)
        _, variable = self.get_sv(module, alias)
        return variable

    def get_sv(self, module, alias):
        def walk(index):
            ml = module.split('.')
            m, vd = '.'.join(ml[:index]), ml[index]
            statement = f'from {m} import {vd}' + s_addition
            ml[index] = alias or ml[index]
            variable = '.'.join(ml[index:])
            return statement, variable, alias or vd

        def conflict(v, s, ss):
            raise ValueError(f'Conflicts with statement: {v} , {s} , {ss}')

        def directly(vv):
            s, v = f'import {module}' + s_addition, module
            ss = self.statements.get(vv)
            if ss is None:
                self.statements[vv] = s
            elif ss != s:
                conflict(v, s, ss)
            return s, v

        s_addition = f' as {alias}' if alias else ''
        s, v, vv = walk(-1)
        if '.' not in module:
            return directly(vv)
        ss = self.statements.get(vv)
        if ss is None:
            self.statements[vv] = s
        elif ss != s:
            s, v, vv = walk(-2)
            ss = self.statements.get(vv)
            if ss is None:
                self.statements[vv] = s
            elif s != s:
                if module.startswith('.'):
                    conflict(v, s, ss)
                vv = alias or module
                return directly(vv)
        return s, v



import functools
from typing import Type
from pathlib import Path
from dektools.str import tab_str
from .template import TemplateWide
from .collection import Collection, DictCollection, NamedListCollection
from .part import Repr, ReprTuple, Cache


class Generator:
    TEMPLATE_DIR = None

    template_name = None
    template_cls = TemplateWide
    template_ext = '.tpl'

    Repr = Repr
    ReprTuple = ReprTuple

    env_default = {}

    def __init__(self, instance, kwargs=None):
        self.instance = instance
        self.parent = None
        self.kwargs = kwargs or {}
        self.cache = Cache(self)

    def render(self):
        raise NotImplementedError

    def render_tpl(self, tpl, variables=None):
        return self.template_cls(self.normalize_variables(
            {**(self.variables if variables is None else variables), **self.kwargs}), self.env_default).render_string(
            str(Path(self.template_path) / (tpl + self.template_ext)))

    def normalize_variables(self, variables):
        return variables

    @property
    @functools.lru_cache(None)
    def template_path(self):
        return str(Path(self.TEMPLATE_DIR) / self.template_name) if self.template_name else str(self.TEMPLATE_DIR)

    @property
    @functools.lru_cache(None)
    def variables(self):
        return self.collect_data('variables', DictCollection)

    def get_variables(self, *args, **kwargs):
        return {
            **{k: self.variables[k] for k in args},
            **{k: self.variables[v] for k, v in kwargs.items()}
        }

    def collect_data(self, entry, collection: Type[Collection]):
        result = collection(entry)
        pre_collect = f'pre_collect_{entry}'
        post_collect = f"post_collect_{entry}"
        if hasattr(self, pre_collect):
            result.set_data(getattr(self, pre_collect)(result.get_data()))
        prefix = f'{entry}_'
        for x in dir(self):
            if x.startswith(prefix):
                name = x[len(prefix):]
                v = getattr(self, x)
                result.append(name, self.cache[x]() if callable(v) else v)
        if hasattr(self, post_collect):
            result.set_data(getattr(self, post_collect)(result.get_data()))
        return result.get_data()

    @staticmethod
    def tab_str(s, n, p=4, sl=False):  # s: list or str
        return tab_str(s, n, p, sl)

    @property
    @functools.lru_cache(None)
    def children(self):
        return self.collect_data('children', NamedListCollection)

    @property
    @functools.lru_cache(None)
    def parent_root(self):
        cursor = self.parent
        while cursor:
            if cursor.parent:
                cursor = cursor.parent
            else:
                return cursor
        return cursor

    def post_collect_children(self, children):
        for lst in children.values():
            for child in lst:
                child.parent = self
        return children

    def post_collect_variables(self, variables):
        children_data = {}
        all_filter_children = getattr(self, f'all_filter_children', lambda k, x, y: x)
        for key, array in self.children.items():
            r = self.tab_str([node.render() for node in array], 0)
            filter_children = getattr(self, f'filter_children_{key}', None)
            if filter_children:
                children_data[key] = filter_children(r, array)
            else:
                children_data[key] = all_filter_children(key, r, array)
        return {
            **variables,
            **children_data,
        }


class GeneratorTpl(Generator):
    template_tpl = 'main'

    def render(self):
        return '' if self.template_tpl is None else self.render_tpl(self.template_tpl)


class GeneratorFiles(Generator):
    def __init__(self, target_dir, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.target_dir = str(target_dir)

    def check(self):
        return self.instance is not None

    def action(self):
        if self.check():
            self.render()
            self.on_rendered()

    def on_rendered(self):
        pass

    def render(self):
        self.template_cls(
            self.normalize_variables({**self.variables, **self.kwargs}),
            env=self.env_default,
            multi=self.multi
        ).render_dir(self.target_dir, self.template_path)
        for generator_cls, ins_list in self.sibling.items():
            for ins in ins_list:
                generator_cls(self.target_dir, ins).action()

    @property
    @functools.lru_cache(None)
    def multi(self):
        return self.collect_data('multi', NamedListCollection)

    @property
    @functools.lru_cache(None)
    def sibling(self):
        return self.collect_data('sibling', DictCollection)



import os
import json
import datetime
from jinja2 import Environment
from jinja2.exceptions import TemplateError
from dektools.common import classproperty
from dektools.file import read_text, read_file, write_file, list_dir, list_relative_path, FileHitChecker, normal_path
from dektools.module import ModuleProxy
from dektools.time import TZ_CURRENT
from dektools.output import obj2str
from .exceptions import RenderException


class ModuleProxyTemplate(ModuleProxy):
    def __call__(self, obj, attrs, *args, **kwargs):
        return self[attrs](obj, *args, **kwargs)

    def __getattr__(self, item):
        if item == 'jinja_pass_arg':
            raise AttributeError(item)
        return super().__getattr__(item)


class Template:
    template_bin_suffix = '-tpl-bin'
    template_suffix = '-tpl'
    target_suffix = None

    file_ignore_name = ''

    _file_ignore_tpl = ['.ignoretpl', '.ignoretpl.tpl']
    _file_ignore_override = ['.ignoreor', '.ignoreor.tpl']
    _file_ignore = ['.ignore', '.ignore.tpl']

    default_env_kwargs = dict(
        autoescape=False
    )

    @classmethod
    def get_file_ignore(cls, name):
        return cls.file_ignore_name + name

    @classproperty
    def file_ignore_tpl(self):
        return [self.get_file_ignore(x) for x in self._file_ignore_tpl]

    @classproperty
    def file_ignore_override(self):
        return [self.get_file_ignore(x) for x in self._file_ignore_override]

    @classproperty
    def file_ignore(self):
        return [self.get_file_ignore(x) for x in self._file_ignore]

    @classproperty
    def files_of_ignore(self):
        return {*self.file_ignore, *self.file_ignore_override, *self.file_ignore_tpl}

    def __init__(self, variables, env=None, filters=None, multi=None, **kwargs):
        self._copy_things = [variables, env, filters, kwargs]
        self.multi = multi or {}
        self.env = Environment(**{
            **self.default_env_kwargs,
            **kwargs
        })
        self.env.filters.update(dict(
            mp=ModuleProxyTemplate()
        ))
        if filters:
            self.env.filters.update(filters)
        self.variables = {
            **{
                'now': datetime.datetime.now(tz=TZ_CURRENT),
                '_virtual_': '__virtual__',
                '_multi_': '__multi__'
            },
            **variables
        }
        if env:
            for k, data in env.items():
                getattr(self.env, k).update(data)

    def copy(
            self,
            variables=None,
            loss_vars=False,
            loss_env=False,
            loss_filters=False,
            loss_multi=False,
            loss_kwargs=False
    ):
        return self.__class__(
            {**({} if loss_vars else self._copy_things[0]), **(variables or {})},
            env=None if loss_env else self._copy_things[1],
            filters=None if loss_filters else self._copy_things[2],
            multi=None if loss_multi else self.multi,
            **({} if loss_kwargs else self._copy_things[3])
        )

    def render(self, content):
        return self.env.from_string(content).render(self.variables)

    @classmethod
    def render_circle(cls, data, **kwargs):
        def to_string(d):
            return json.dumps(d, sort_keys=True)

        def to_data(s):
            return json.loads(s)

        cursor_data = data
        cursor_str = to_string(cursor_data)
        while True:
            prev_str = cursor_str
            cursor_str = cls(cursor_data, **kwargs).render(cursor_str)
            if cursor_str == prev_str:
                return cursor_data
            cursor_data = to_data(cursor_str)

    def render_string(self, template_file, close_tpl=False):
        if close_tpl:
            return read_file(template_file)
        else:
            content = read_text(template_file)
            try:
                return self.render(content)
            except TemplateError as e:
                raise RenderException(f'Render error: {template_file}: {e.args} ==>:\n{obj2str(self.variables)}')

    def render_file(self, target_file, template_file, close_tpl=False):
        content = self.render_string(template_file, close_tpl)
        write_file(target_file, sb=content)
        return content

    def render_dir(self, path_target, path_template, force_close_tpl=False, open_ignore_override=True):
        files_of_ignore = self.files_of_ignore
        match_ignore_tpl = FileHitChecker(
            path_template,
            self.file_ignore_tpl[0],
            lines=self.get_hit_rules(path_template, self.file_ignore_tpl[1])
        ).new_match()
        match_ignore_override = FileHitChecker(
            path_template,
            self.file_ignore_override[0],
            lines=self.get_hit_rules(path_template, self.file_ignore_override[1])
        ).new_match()
        match_ignore = FileHitChecker(
            path_template,
            self.file_ignore[0],
            lines=[
                *self.get_hit_rules(path_template, self.file_ignore[1]),
                *[f'/{item}' for item in [*self.file_ignore_tpl, *self.file_ignore_override, *self.file_ignore]]
            ]
        ).new_match()

        path_virtual = os.path.join(path_template, '__virtual__')
        variables_virtual = {}
        for rp, fp in list_relative_path(path_virtual).items():
            variables_virtual[rp.replace('\\', '/').replace('/', '__')] = self.render_string(fp)
        self.variables['__virtual__'] = variables_virtual

        path_multi = os.path.join(path_template, '__multi__')
        for fp, name in list_dir(path_multi, True):
            list_variables = self.multi.get(name)
            if list_variables:
                for variables in list_variables:
                    tmpl = self.copy(variables, loss_vars=True, loss_multi=True)
                    tmpl.render_dir(path_target, os.path.join(path_multi, name), force_close_tpl, open_ignore_override)

        for root, _, files in os.walk(path_template):
            if any(root.startswith(x + os.path.sep) for x in [path_virtual, path_multi]):
                continue
            for f in files:
                if f in files_of_ignore and root == path_template:
                    continue
                fp = os.path.join(root, f)
                if match_ignore(fp):
                    continue
                fp_ = fp
                is_bin = False
                if fp_.endswith(self.template_bin_suffix):
                    fp_ = fp_[:len(fp_) - len(self.template_bin_suffix)]
                    is_bin = True
                elif fp_.endswith(self.template_suffix):
                    fp_ = fp_[:len(fp_) - len(self.template_suffix)]
                rp = fp_[len(path_template):]
                target_file = path_target + self.render(rp)
                if open_ignore_override and os.path.exists(target_file) and match_ignore_override(fp):
                    continue
                if is_bin:
                    write_file(target_file, sb=self.variables[self.render(read_text(fp)).strip()])
                else:
                    self.render_file(target_file, fp, force_close_tpl or match_ignore_tpl(fp))

    def get_hit_rules(self, path_template, filepath):
        ignore_file = normal_path(os.path.join(path_template, filepath))
        if os.path.exists(ignore_file):
            content = self.render_string(ignore_file)
            return [line for line in content.split('\n') if line]
        else:
            return []


class TemplateWide(Template):
    default_env_kwargs = dict(
        **Template.default_env_kwargs,
        variable_start_string='(=(',
        variable_end_string=')=)'
    )



import functools
from .imports import Imports, ImportsStatement
from .template import TemplateFormatted
from .template.env import env_extend
from ...tmpl.generator import GeneratorTpl, GeneratorFiles, Repr, DictCollection


class ReprTS(Repr):
    def final_value(self):
        return f'_({repr(self.value)})'


class CodeGeneratorMixin:
    ReprTS = ReprTS
    template_cls = TemplateFormatted
    _imports_cls = Imports
    _imports_statement_cls = ImportsStatement

    env_default = {**env_extend.context}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__imports_statement__ = None

    @property
    @functools.lru_cache(None)
    def iss(self):
        if self.parent_root:
            return self.parent_root.__imports_statement__
        else:
            _imports_core = getattr(self.__class__, '_imports_core', None)
            if _imports_core is None:
                _imports_core = self.__class__._imports_core = self._imports_cls()
            imports_core = _imports_core.derive()
            imports_core.update(self.walk_imports(self))
            self.__imports_statement__ = self._imports_statement_cls(imports_core)
            return self.__imports_statement__

    def post_collect_variables(self, variables):
        variables = super().post_collect_variables(variables)
        if self.parent_root:
            imports_data = {}
        else:
            imports_data = dict(
                imports=self.iss.as_statements()
            )
        return {
            **variables,
            **imports_data
        }

    @property
    @functools.lru_cache(None)
    def imports(self):
        return self.collect_data('imports', DictCollection)

    @classmethod
    def walk_imports(cls, node):
        imports = {}
        imports.update(node.imports)
        for children in node.children.values():
            for child in children:
                imports.update(cls.walk_imports(child))
        return imports


class CodeGeneratorTpl(CodeGeneratorMixin, GeneratorTpl):
    pass


class CodeGeneratorFiles(CodeGeneratorMixin, GeneratorFiles):
    pass



import os
import black
from black.parsing import InvalidInput
from dektools.file.operation import write_file
from ....tmpl.template import TemplateWide


class TemplateException(Exception):
    pass


class FormatException(Exception):
    pass


def py_formatted(s):
    if not isinstance(s, str):
        s = s.decode('utf-8')
    try:
        return black.format_str(s, mode=black.Mode(line_length=120, string_normalization=False))
    except InvalidInput:
        raise FormatException()


class TemplateFormatted(TemplateWide):
    format_ext_map = {
        '.py': py_formatted
    }

    def render_file(self, target_file, template_file, close_tpl=False):
        content = self.render_string(template_file, close_tpl)
        formatted = self.format_ext_map.get(os.path.splitext(target_file)[-1], default_formatted)
        try:
            content = formatted(content)
            errors = False
        except FormatException:
            errors = True
        write_file(target_file, sb=content)
        if errors:
            raise TemplateException(f'format error, please check the target file: {template_file} , {target_file}')
        return content


def default_formatted(s):
    return s



import os
from collections import OrderedDict
from dektools.zip import decompress_files
from dektools.file import FileHitChecker, normal_path
from ..utils.yaml import yaml, data_final_list
from ..tmpl.template import TemplateWide


class ResLoader:
    accept_ext_set = {'.yaml'}
    ignore_prefix = '_'
    ignore_res = '.resignore'

    template_cls = TemplateWide
    serializer = yaml
    data_final_list = data_final_list

    @classmethod
    def data_final(cls, data_list):
        for df in data_final_list:
            data_list = df(data_list)
        return data_list

    @classmethod
    def is_res_file(cls, filepath):
        return bool(cls.get_res_name(filepath))

    @classmethod
    def get_res_name(cls, filepath):
        filename = os.path.basename(filepath)
        fb, ext = os.path.splitext(filename)
        if ext in cls.accept_ext_set and not fb.startswith(cls.ignore_prefix):
            return fb
        return None

    @classmethod
    def glob_files(cls, filepath, judge, *ignores):
        result = []
        if os.path.isfile(filepath):
            result.append(filepath)
        elif os.path.isdir(filepath):
            for f in os.listdir(filepath):
                result.extend(cls.glob_files(os.path.join(filepath, f), None))
        if ignores:
            fhc = FileHitChecker(filepath, *ignores)
            return [x for x in result if not judge or judge(cls.is_res_file, fhc.new_match(), x)]
        else:
            return result

    @classmethod
    def glob_res_file(cls, filepath, ignore=ignore_res):
        return cls.glob_files(filepath, lambda is_res_file, is_hit, x: is_res_file(x) and not is_hit(x), ignore)

    @classmethod
    def glob_not_res_file(cls, filepath, ignore=ignore_res):
        return cls.glob_files(filepath, lambda is_res_file, is_hit, x: not is_res_file(x) or is_hit(x), ignore)

    @classmethod
    def data_from_file(cls, close_tpl, *args):
        data_list = []
        for arg in cls._extend_filepath_args(args):
            data = cls.serializer.loads(
                cls.template_cls(
                    arg.get('variables') or {}, **arg.get('kwargs') or {}
                ).render_string(arg['filepath'], close_tpl=close_tpl)
            ) or OrderedDict()
            if 'order' in arg:
                data['__order__'] = arg['order']
            data['__file__'] = arg['filepath']
            data['__context__'] = arg.get('context') or {}
            data_list.append(data)
        data_list = sorted(data_list, key=lambda x: x.get('__order__') or 0)
        return cls.data_final(data_list)

    @classmethod
    def data_from_template_file(cls, *args):
        return cls.data_from_file(False, *args)

    @classmethod
    def from_template_file(cls, *args):
        return cls(cls.data_from_template_file(*args))

    @classmethod
    def load_res(cls, path_res, close_tpl=False, path=None):
        if not os.path.exists(path_res):
            raise FileNotFoundError(path_res)
        if os.path.isfile(path_res):
            path_res = decompress_files(path_res)
        path_res = normal_path(path_res)
        if path:
            path_res = path(path_res)
        return cls(cls.data_from_file(close_tpl, *(dict(filepath=x) for x in cls.glob_res_file(path_res))))

    @classmethod
    def _extend_filepath_args(cls, args):
        def get_key_map(pa):
            r = {}
            for path, arg in pa.items():
                key = get_path_unique(path)
                if key not in r or r[key][1] > len(path):
                    r[key] = [arg, len(path)]
            return {key: v[0] for key, v in r.items()}

        def get_path_unique(path):
            file_base = os.path.basename(path).split('.', 1)[0]
            dir_path = os.path.dirname(path)
            return dir_path, file_base

        def extend_to_valid_paths(paths):
            path_children = {}
            path_extends = OrderedDict()
            for path in paths:
                key = get_path_unique(path)
                dir_path, file_base = key
                if dir_path not in path_children:
                    path_children[dir_path] = sorted(os.listdir(dir_path))
                if key not in path_extends:
                    path_list = path_extends[key] = []
                    for fn in path_children[dir_path]:
                        fp = os.path.join(dir_path, fn)
                        if cls.is_res_file(fp) and fn.startswith(f'{file_base}.'):
                            path_list.append(fp)
            return path_extends

        path_map = OrderedDict([(os.path.normpath(os.path.abspath(arg['filepath'])), arg) for arg in args])
        key_map = get_key_map(path_map)
        result = []
        for k, fl in extend_to_valid_paths(path_map.keys()).items():
            for f in fl:
                if f in path_map:
                    arg = path_map[f]
                else:
                    arg = {**key_map[k], **dict(filepath=f)}
                result.append(arg)
        return result

    def __init__(self, data_list):
        self.data_list = data_list



from copy import deepcopy
from itertools import chain
from dektools.str import str_special_escape
from dektools.dict import is_list, is_dict, merge_flex, assign_flex
from dektools.yaml.tags.base import TagBase
from ....tmpl.template import Template


class TemplateTmpl(Template):
    default_env_kwargs = dict(
        **Template.default_env_kwargs,
        variable_start_string='(-(',
        variable_end_string=')-)'
    )


class TmplBase:
    sep = '/'

    @classmethod
    def parse_args(cls, array):
        return dict()

    @classmethod
    def pkg(cls, exp, data):
        array = exp.lstrip(cls.sep).split(cls.sep) if exp else []
        return cls(**{'data': data, **cls.parse_args(array)})

    def __init__(self, data=None):
        self.data = data


class TmplPatch(TmplBase):
    sep_operation = ':'

    @classmethod
    def parse_args(cls, array):
        func = array[0]
        args = array[1:]
        array = func.split(cls.sep_operation)
        operation = None
        if len(array) > 1:
            operation = array[1]
        func = array[0]
        return dict(func=func, operation=operation, args=args)

    def __init__(self, func=None, operation=None, args=None, **kwargs):
        super().__init__(**kwargs)
        self.args = args
        self.func = func
        self.operation = operation  # merge_flex: _merge_options

    @staticmethod
    def walk_data(node, walk):
        if is_list(node):
            for key, value in enumerate(node):
                walk(node, key, value)
        elif is_dict(node):
            for key, value in node.items():
                walk(node, key, value)

    def combine_data(self, func):
        data = deepcopy(func.data)
        if not self.data:
            return data
        data2 = deepcopy(self.data)
        merge_flex(data, data2, self.operation)
        return data

    def invoke(self, ins, key, func):
        raise NotImplementedError()


class FuncTmpl(TmplBase):
    @classmethod
    def parse_args(cls, array):
        return dict(args=array)

    def __init__(self, args=None, **kwargs):
        super().__init__(**kwargs)
        self.args = args

    def get_context(self, args):
        context = {}
        for i, k in enumerate(self.args):
            kv = k.split('=')
            if len(kv) == 1:
                context[k] = args[i]
            else:
                k, v = kv
                if i < len(args):
                    v = args[i]
                context[k] = v
        return context


class CallTmpl(TmplPatch):
    def invoke(self, ins, key, func):
        if isinstance(key, str) and key.startswith('='):
            ins.pop(key)
            ins.update(self.invoke_inner(func))
        else:
            ins[str_special_escape(key) if isinstance(key, str) else key] = self.invoke_inner(func)

    def invoke_inner(self, func):
        def walk(node, key, value):
            self.walk_data(value, walk)
            if isinstance(key, str) or isinstance(value, str):
                replace_list.append(delay(node, key, value))

        def delay(ins, key, value):
            def wrapper():
                nonlocal key, value
                if isinstance(key, str):
                    ins.pop(key)
                    key = tt.render(key)
                if isinstance(value, str):
                    value = tt.render(value)
                ins[key] = value

            return wrapper

        tt = TemplateTmpl(func.get_context(self.args))
        replace_list = []
        data_total = self.combine_data(func)
        self.walk_data(data_total, walk)
        for c in replace_list:
            c()
        return data_total


class ExtendTmpl(TmplPatch):
    def invoke(self, ins, key, func):
        data = self.invoke_inner(func)
        if is_list(ins) or not is_dict(data):
            ins[key] = data
        else:
            temp = ins.__class__()
            combine = False
            for k, v in ins.items():
                if k == key:
                    combine = True
                    v = data
                if combine:
                    temp = assign_flex(temp, v)
                else:
                    temp[k] = v
            ins.clear()
            ins.update(temp)

    def invoke_inner(self, func):
        return self.combine_data(func)


def tmpl_data_final(data):
    def walk(node):
        if is_list(node):
            for key, value in enumerate(node):
                walk2(node, key, value)
        elif is_dict(node):
            for key, value in node.items():
                walk2(node, key, value)

    def walk2(node, key, value):
        walk(value)
        if isinstance(value, FuncTmpl):
            walk(value.data)
            func_map[key] = value
            func_list.append(lambda: node.pop(key))
        elif isinstance(value, ExtendTmpl):
            walk(value.data)
            extend_list.append(delay_patch(node, key, value))
        elif isinstance(value, CallTmpl):
            walk(value.data)
            call_list.append(delay_patch(node, key, value))

    def delay_patch(ins, key, patch):
        def wrapper():
            patch.invoke(ins, key, func_map[patch.func])

        return wrapper

    func_map = {}
    func_list = []
    extend_list = []
    call_list = []
    walk(data)
    for c in chain(func_list, extend_list, call_list):
        c()
    return data


class TmplTag(TagBase):
    yaml_tag_multi = True
    tmpl_cls = None

    @classmethod
    def from_yaml_multi(cls, loader, tag_suffix, node):
        return cls.tmpl_cls.pkg(tag_suffix, cls.node_to_data(loader, node))


class FuncTag(TmplTag):
    yaml_tag = '!f'
    tmpl_cls = FuncTmpl


class CallTag(TmplTag):
    yaml_tag = '!c'
    tmpl_cls = CallTmpl


class ExtendTag(TmplTag):
    yaml_tag = '!e'
    tmpl_cls = ExtendTmpl



import os
from dektools.yaml import yaml
from dektools.dict import dict_merge
from dektools.typer import load_yaml_files
from .template import TemplateWide


class RenderTemplate(TemplateWide):
    file_ignore_tpl = [f"../{x}" for x in TemplateWide.file_ignore_tpl]
    file_ignore_override = [f"../{x}" for x in TemplateWide.file_ignore_override]
    file_ignore = [f"../{x}" for x in TemplateWide.file_ignore]


def render_path(dest, src, files=None, updated=None, **kwargs):
    data = {}
    path_values = os.path.join(src, 'values.yaml')
    if os.path.isfile(path_values):
        data = dict(Values=yaml.load(path_values) or {})
    dict_merge(data, dict(Values=load_yaml_files(files)))
    if updated:
        dict_merge(data, updated)
    if os.path.isdir(dest) or os.path.isdir(src):
        RenderTemplate(data, **kwargs).render_dir(dest, os.path.join(src, 'templates'))
    elif os.path.isfile(src):
        RenderTemplate(data, **kwargs).render_file(dest, src)



import functools


class Cache:
    def __init__(self, master):
        self.master = master
        self.caches = {}

    def __getitem__(self, method):
        return functools.partial(self.__run, method)

    def __getattr__(self, method):
        return self[method]

    def __run(self, method):
        if method not in self.caches:
            self.caches[method] = getattr(self.master, method)()
        return self.caches[method]


class Repr:
    def __init__(self, value, delay=False):
        self.value = value
        self.delay = delay

    def final_value(self):
        return self.value

    def __repr__(self):
        value = self.final_value()
        return f'lambda: {value}' if self.delay else value

    def __copy__(self):
        return self.__class__(self.value, self.delay)

    def __deepcopy__(self, memo):
        return self.__class__(self.value, self.delay)


class ReprTuple(Repr):
    def __init__(self, value, delay=False):
        super().__init__(list(value), delay)

    def final_value(self):
        return repr(tuple(self.value))

    def __getattr__(self, item):
        return getattr(self.value, item)



class EnvExtend:
    def __init__(self):
        self._filters = {}
        self._globals = {}

    @staticmethod
    def mark(context, func_or_name):
        def func_wrapper(func):
            context[func_or_name] = func
            return func

        if isinstance(func_or_name, str):
            return func_wrapper
        else:
            context[func_or_name.__name__] = func_or_name
            return func_or_name

    def filters(self, func_or_name):
        return self.mark(self._filters, func_or_name)

    def globals(self, func_or_name):
        return self.mark(self._globals, func_or_name)

    @property
    def context(self):
        return dict(
            filters=self._filters,
            globals=self._globals
        )



from dektools.dict import is_list


class Collection:
    def __init__(self, entry):
        self.entry = entry

    def append(self, name, data):
        raise NotImplementedError

    def set_data(self, data):
        raise NotImplementedError

    def get_data(self):
        raise NotImplementedError


class CollectionGeneric(Collection):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data = {}

    def set_data(self, data):
        self.data = data

    def get_data(self):
        return self.data


class DictCollection(CollectionGeneric):
    def append(self, name, data):
        self.data.update(data)


class NamedCollection(CollectionGeneric):
    def append(self, name, data):
        self.data[name] = data


class NamedListCollection(NamedCollection):
    def append(self, name, data):
        if is_list(data):
            super().append(name, data)
        else:
            self.data.update(data)



from dektools.yaml import yaml as yaml_, Yaml, Resolver
from .tags import FuncTag, CallTag, ExtendTag, tmpl_data_final

yaml = Yaml(resolvers=Resolver.ordereddict)
yaml.reg_other(yaml_)

yaml.reg_batch(FuncTag, CallTag, ExtendTag)

data_final_list = [tmpl_data_final]



import typer
from typing import Optional, List
from typing_extensions import Annotated
from dektools.typer import multi_options_to_dict
from ..tmpl.render import render_path

app = typer.Typer(add_completion=False)


@app.command()
def render(
        dest,
        src,
        vss=None,
        ves=None,
        files: Annotated[Optional[List[str]], typer.Option('--file')] = None,
        _set: Annotated[Optional[List[str]], typer.Option('--set')] = None
):
    if vss and ves:
        kwargs = dict(
            variable_start_string=vss,
            variable_end_string=ves
        )
    else:
        kwargs = {}
    render_path(dest, src, files, multi_options_to_dict(_set), **kwargs)


@app.callback()
def callback():
    pass
